module Transpiler
using CUDA
using ..ExpressionProcessing

# culoadtest(N, rand(["add.f32", "sub.f32", "mul.f32", "div.approx.f32"]))
function culoadtest(N::Int32, op = "add.f32")

	vadd_code = ".version 7.1
	
	.target sm_52
	.address_size 64
	
		// .globl       VecAdd_kernel
	
	.visible .entry VecAdd_kernel(
		.param .u64 VecAdd_kernel_param_0,
		.param .u64 VecAdd_kernel_param_1,
		.param .u64 VecAdd_kernel_param_2,
		.param .u32 VecAdd_kernel_param_3
	)
	
	{
		.reg .pred      %p<2>;
		.reg .f32       %f<4>;
		.reg .b32       %r<6>;
		.reg .b64       %rd<11>;

		ld.param.u64    %rd1, [VecAdd_kernel_param_0];
		ld.param.u64    %rd2, [VecAdd_kernel_param_1];
		ld.param.u64    %rd3, [VecAdd_kernel_param_2];
		ld.param.u32    %r2, [VecAdd_kernel_param_3];

		mov.u32         %r3, %ntid.x;
		mov.u32         %r4, %ctaid.x;
		mov.u32         %r5, %tid.x;

		mad.lo.s32      %r1, %r3, %r4, %r5;

		setp.ge.s32     %p1, %r1, %r2;

		@%p1 bra        \$L__BB0_2;

		cvta.to.global.u64      %rd4, %rd1;

		mul.wide.s32    %rd5, %r1, 4;
		add.s64         %rd6, %rd4, %rd5;
		cvta.to.global.u64      %rd7, %rd2;
		add.s64         %rd8, %rd7, %rd5;

		ld.global.f32   %f1, [%rd8];
		ld.global.f32   %f2, [%rd6];" *
		op *
		"               %f3, %f2, %f1;
		cvta.to.global.u64      %rd9, %rd3;
		add.s64         %rd10, %rd9, %rd5;
		st.global.f32   [%rd10], %f3;
	
	\$L__BB0_2:
		ret;
	}"
	
		linker = CuLink()
		add_data!(linker, "VecAdd_kernel", vadd_code)
	
		image = complete(linker)
	
		mod = CuModule(image)
		func = CuFunction(mod, "VecAdd_kernel")
	
		d_a = CUDA.fill(1.0f0, N)
		d_b = CUDA.fill(2.0f0, N)
		d_c = CUDA.fill(0.0f0, N)
	
		# Grid/Block configuration
		threadsPerBlock = 256;
		blocksPerGrid  = (N + threadsPerBlock - 1) ÷ threadsPerBlock;
	
		@time CUDA.@sync cudacall(func, Tuple{CuPtr{Cfloat},CuPtr{Cfloat},CuPtr{Cfloat},Cint}, d_a, d_b, d_c, N; threads=threadsPerBlock, blocks=blocksPerGrid)
end

# Number of threads per block/SM + max number of registers
# https://docs.nvidia.com/cuda/cuda-c-programming-guide/#features-and-technical-specifications
# Need to assume a max of 2048 threads per Streaming Multiprocessor (SM)
# One SM can have 64*1024 32-bit registers at max
# One thread can at max use 255 registers
# Meaning one has access to at most 32 registers in the worst case. Using 64 bit values this number gets halfed (see: https://docs.nvidia.com/cuda/cuda-c-programming-guide/#multiprocessor-level (almost at the end of the linked section))

# I think I will go with max 16 registers for now and leave a better register allocation technique for future work
# Maybe helpful for future tuning: https://docs.nvidia.com/cuda/cuda-c-programming-guide/#maximum-number-of-registers-per-thread

# https://docs.nvidia.com/cuda/cuda-c-programming-guide/#multiprocessor-level
# This states, that using fewer registers allows more threads to reside on a single SM which improves performance. 
# So I could use more registers at the expense for performance. Depending on how this would simplify my algorithm, I might do this and leave more optimisation to future work

# Since the generated expressions should have between 10 and 50 symbols, I think allowing a max. of 128 32-bit registers should make for an easy algorithm. If during testing the result is slow, maybe try reducing the number of registers and perform more intelligent allocation/assignment
# With 128 Registers, one could have 32 Warps on one SM ((128 * 16 = 2048) * 32 == 64*1024 == max number of registers per SM) This means 512 Threads per SM in the worst case

#
# Make a "function execute(...)" that takes the data and the transpiled code. Pass the data to the kernel and start executing
# Note: Maybe make an additional function that transpiles and executed the code. This would then be the function the user calls
#

# TODO: I guess I need to move this to different module because it seems like I can't access inner functions
let registers = Dict() # stores the count of the register already used.
	global get_next_free_register
	global get_used_registers

	# By convention these names correspond to the following types:
	# - p -> pred
	# - f32 -> float32
	# - b32 -> 32 bit
	# - var -> float32
	# - param -> float32 !! although, they might get inserted as fixed number and not be sent to gpu?
	function get_next_free_register(name::String)::String
		if haskey(registers, name)
			registers[name] += 1
		else
			registers[name] = 1
		end

		return String(["%", name, registers[name]])
	end

	function get_used_registers()
		return pairs(registers)
	end
end

# To increase performance, it would probably be best for all helper functions to return their IO Buffer and not a string
const exitJumpLocationMarker = "\$L__BB0_2"
function transpile(expression::ExpressionProcessing.PostfixType)::String
	ptxBuffer = IOBuffer()

	println(ptxBuffer, get_cuda_header())
	println(ptxBuffer, get_kernel_signature("ExpressionProcessing", [Int32, Float32]))
	println(ptxBuffer, "{")

	# TODO: Actually calculate the number of needed registers and extend to more register kinds
	println(ptxBuffer, get_register_definitions(1, 5, 0))  # apparently I can define registers anywhere. This might make things easier
	# TODO: Parameter loading
	println(ptxBuffer, get_guard_clause())

	# top down create the code. keep track of the max number of variables/parameters used (needed for later iterations. See section "Plan" in "PTX_understanding.md")
	# return this alongside the generated code
	# Generate registers based off of the above number
	# Variables have: %var0 to %varn - 1
	# Parameters have: %param0 to %paramn - 1
	# Code goes here
	(calc_code, fRegisterCount) = generate_calculation_code(expression)
	println(ptxBuffer, get_register_definitions(0, 0, fRegisterCount))
	println(ptxBuffer, calc_code)

	# exit jump location
	print(ptxBuffer, exitJumpLocationMarker); println(ptxBuffer, ": ret;")
	println(ptxBuffer, "}")

	generatedCode = String(take!(ptxBuffer))
	println(generatedCode)
	return generatedCode
end

# TODO: Make version, target and address_size configurable; also see what address_size means exactly
function get_cuda_header()::String
	return "
.version 7.1
.target sm_52
.address_size 64
"
end

function get_kernel_signature(kernelName::String, parameters::Vector{DataType})::String
	signatureBuffer = IOBuffer()
	print(signatureBuffer, ".visible .entry ")
	print(signatureBuffer, kernelName)
	println(signatureBuffer, "(")

	
	for i in eachindex(parameters)
		type = type_to_ptx_type(parameters[i])
		print(signatureBuffer, 
			  "  .param ", type, " ", kernelName, "_param_", i)
		if i != lastindex(parameters)
			println(signatureBuffer, ",")
		end
	end

	print(signatureBuffer, ")")
	return String(take!(signatureBuffer))
end

"
Constructs the PTX code used for handling the case where too many threads are started.

Assumes the following:
 - There are the unused ```32 bit``` registers ```r0, r1, r2, r3 (index of the variable set)```
 - There is an unused ```predicate``` register ```p0```
 - The ```32 bit``` register ```r4``` contains the number of variable sets
"
function get_guard_clause()::String
	guardBuffer = IOBuffer()

	println(guardBuffer, "mov.u32    %r0, %ntid.x;") # nr of thread ids
	println(guardBuffer, "mov.u32    %r1, %ctaid.x;") # nr of threads per cta
	println(guardBuffer, "mov.u32    %r2, %tid.x;") # id of the current thread

	println(guardBuffer, "mad.lo.s32    %r3, %r0, %r1, %r2;") # the current index (basically index of variable set)
	println(guardBuffer, "setp.ge.s32    %p0, %r3, %r4;") # guard clause (p0 = r3 > r4 -> index > nrOfVariableSets)

	# branch to end if p0 is true
	print(guardBuffer, "@%p0 bra    $exitJumpLocationMarker;")

	return String(take!(guardBuffer))
end

# TODO: refactor this for better usage. Maybe make this generate only one register definition and pass in the details
function get_register_definitions(nrPred::Int, nr32Bit::Int, nrFloat32::Int)::String
	registersBuffer = IOBuffer()

	if nrPred > 0
		println(registersBuffer, ".reg .pred   %p<$nrPred>;")
	end
	if nr32Bit > 0
		println(registersBuffer, ".reg .b32    %r<$nr32Bit>;")
	end
	if nrFloat32 > 0
		println(registersBuffer, ".reg .f32    %f<$nrFloat32>;")
	end

	return String(take!(registersBuffer))
end

# Current assumption: Expression only made out of constant values
function generate_calculation_code(expression::ExpressionProcessing.PostfixType)::Tuple{String, Int}
	codeBuffer = IOBuffer()
	operands = Vector{Union{Float32, String}}() # Maybe make it of type ANY. Then I could put the register name "on the stack" instead and build up the code like that. Could also make it easier implementing variables/params

	registerCounter = 0
	println(expression)
	for i in eachindex(expression)
		token = expression[i]

		if token.Type == FLOAT32
			push!(operands, reinterpret(Float32, token.Value))
		elseif token.Type == OPERATOR
			# function call to see if operator is unary -> adapt below calculation; probably able to reuse register
			operator = get_ptx_operator(reinterpret(Operator, token.Value))
			register = "%f$registerCounter"
			print(codeBuffer, "    $operator $register, ")

			ops = last(operands, 2)
			pop!(operands);pop!(operands)
			print(codeBuffer, join(ops, ", "))
			println(codeBuffer, ";")

			push!(operands, register)
			registerCounter += 1
		elseif token.Type == INDEX
			# TODO
		end
	end

	return (String(take!(codeBuffer)), registerCounter)
end

function type_to_ptx_type(type::DataType)::String
	if type == Int64
		return ".s64"
	elseif type == Float32
		return ".f32"
	else
		return ".b64"
	end
end

# TODO: Probably change this, to return the entire calculation not just the operator. Because for POWER and EXP we need multiple instructions to calculate them (seperation of concerns).
function get_ptx_operator(operator::Operator)::String
	if operator == ADD
		return "add.f32"
	elseif operator == SUBTRACT
		return "sub.f32"
	elseif operator == MULTIPLY
		return "mul.f32"
	elseif operator == DIVIDE
		return "div.approx.f32"
	elseif operator == POWER
		return ""
	elseif operator == ABS
		return "abs.f32"
	elseif operator == LOG
		return "lg2.approx.f32"
	elseif operator == EXP
		return ""
	elseif operator == SQRT
		return "sqrt.approx.f32"
	else
		throw(ArgumentError("Operator conversion to ptx not implemented for $operator"))
	end
end



end

